td.number {
	text-align: right;
	font-weight: bold;
	padding-right: 5px;
	white-space: nowrap;
}

td.number.nan {
	font-weight: normal;
	color: #AAA;
}

th.number {
	text-align: right;
}

td.boolean {
	text-align: center;
}

th.boolean {
	text-align: center;
}
