body {
	font-family:'lucida grande', tahoma, verdana, arial, sans-serif;
	font-size:11px;
}

#wrap {
	width: 850px;
}

h1 { 
	font-size: 15px;
}

h2 { 
	font-size: 13px;
}

a { 
	color: #548dc4;
	text-decoration: none;
}

h1 a {
	font-size: 12px;
	float: right;
}

a:hover {
	text-decoration: underline;
}

img {
	border: 0;
}

.ok { 
	background-color: #0a8819; 
	color: white; 
	padding: 5px;
}

input.invalid {
	background: red;
	color: #FDFDFD;
}

#message {
	width: 100%;
}

table.testgrid { 
	border-collapse: collapse; 
	border: 1px solid #CCB;
	width: 100%;
}

table.testgrid tr {
	border-bottom: 1px solid #DDD;
}

table.testgrid>tbody tr:hover {
	background-color: #888888;
	color: white;
}

table.testgrid>tbody tr:hover a {
	color: white;
}

table.testgrid th {
	background: #E5E5E5;
	border: 1px solid #D5D5D5;
	color: #555;
	text-align: left;
	padding-left: 5px;
	padding-right: 0px;
	padding-top: 5px;
	padding-bottom: 5px;
	white-space: nowrap; 
}

table.testgrid td {
	padding: 5px;
	border: 1px solid #E0E0E0;
}